;;; Copyright (C) 2024, 2025 Igalia, S.L.
;;;
;;; Licensed under the Apache License, Version 2.0 (the "License");
;;; you may not use this file except in compliance with the License.
;;; You may obtain a copy of the License at
;;;
;;;    http://www.apache.org/licenses/LICENSE-2.0
;;;
;;; Unless required by applicable law or agreed to in writing, software
;;; distributed under the License is distributed on an "AS IS" BASIS,
;;; WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
;;; See the License for the specific language governing permissions and
;;; limitations under the License.

(use-modules (srfi srfi-64)
             (test utils))

(test-begin "test-expander")

(with-imports ((hoot tree-il)
               (hoot modules)
               (hoot expander)
               ;; FIXME: error-handling and exceptions needed to
               ;; initialize $make-match-error, $raise-exception et al.
               (hoot error-handling)
               (hoot exceptions)
               (hoot syntax))

 (test-call "42"
            (lambda (x)
              (let ((m (make-empty-module)))
                (const-exp (expand-syntax x m))))
            #'42))

(test-end* "test-expander")
