// Hoot read-eval-print loop
// Copyright (C) 2026 David Thompson <dave@spritely.institute>
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

const [ reflectJsDir, reflectWasmDir, replWasm, ...loadPath ] = process.argv.slice(3);
const { Scheme, SchemeQuitError, repr } = await import(`${reflectJsDir}/reflect.js`);
try {
  const [proc] = await Scheme.load_main(replWasm, {
    reflect_wasm_dir: reflectWasmDir
  });
  await proc.call_async(...loadPath);
} catch(e) {
  if (e instanceof SchemeQuitError) {
    process.exit(e.status);
  } else {
    console.log(`error: ${e} (${e.stack})`);
    process.exit(1);
  }
}
process.exit(0);
